#! /usr/bin/python3

"""
     __
  /\ \ \_ __  _   _  __ _  __ _  __ _
 /  \/ / '_ \| | | |/ _` |/ _` |/ _` |
/ /\  /| | | | |_| | (_| | (_| | (_| |
\_\ \/ |_| |_|\__, |\__,_|\__,_|\__,_|
    althonos  |___/  the Nth NYAA API

"""

__version__ = '1.0.1'

import re
import os
import sys
import multiprocessing
import dateutil.parser
import argparse

try:
    import urllib.request as rqst
except ImportError:
    import urllib2 as rqst
from bs4 import BeautifulSoup

#const
URL = "http://www.nyaa.se"

#conditional color wrappers
green  = lambda x, boole: "\033[92m"*boole+x+"\033[39m"*boole
red    = lambda x, boole: "\033[91m"*boole+x+"\033[39m"*boole
cyan   = lambda x, boole: "\033[36m"*boole+x+"\033[39m"*boole
bold   = lambda x, boole: "\033[1m"*boole +x+"\033[21m"*boole

#settings dict
typecolor_dict = {'std':lambda x,c:x, 'trs':green, 'rmk':red, 'a+':cyan}
sort_dict =  {'date':1, 'seeders':2, 'leechers':3, 'downloads':4, 'size':5}
order_dict = {'descending':1, 'ascending':2, 'desc':1, 'asc':2}
filter_dict = {'all':0, 'noremake':1, 'trusted':2, 'aplus':3}



class Downloader(multiprocessing.Process):
    """
    A standalone downloading Process.
    Downloader.join() can be run, however it is not ran inside of the Nnyaaa API
    as the torrent size are small enough to suppose download will be over at the
    end of the code execution.
    """

    def __init__(self, dl_url, dl_dir, dl_name, headers={'User-Agent' : 'Firefox'}):
        super().__init__()
        self.dl_url = dl_url
        self.dl_dir = dl_dir
        self.dl_name = dl_name
        self.headers = headers

    def run(self):
        if not os.path.isdir(self.dl_dir):
            os.mkdir(self.dl_dir)
        with open(os.path.join(self.dl_dir, self.dl_name), 'wb') as f:
            req = rqst.Request(self.dl_url, headers=self.headers)
            con = rqst.urlopen(req)
            f.write(con.read())


class TList:
    """
    The representation of a result list (search page or main page).
    Facultative Arguments:
        - query (default=None)      a string containing the query to search for.
                                    if query is None, then the TList will display
                                    the results of nyaa main page.
        - sort (default='date')     a string containing the sorting form
                                    (choice in 'date', 'seeders', 'leechers',
                                    'downloads', 'size')
        - order(default='desc')     a string containing the order of sorting
                                    (choice in 'asc', 'desc')
        - filter(default = 'all')   a string containing filter query provided by
                                    Nyaa (choice in 'all', 'noremake',
                                    'trusted', 'aplus')
        - start_offset (default=1)  integer containing the number of the page to
                                    start looking at
        - max_offset (default=1)    an integer containing the greatest page to
                                    parse. Set to -1 to parse all pages
                                    (! don't do that on large query sets !)
    """

    def __init__(self, query=None, sort='date', order='desc', filter='all',
                 start_page=1, end_page=1, color=False):
        self.query = query
        self.color = color
        self.sort = sort_dict[sort]
        self.order = order_dict[order]
        self.max_offset = end_page
        self.start_offset = start_page
        self.filter = filter_dict[filter]

    def build_url(self):
        self.url = URL + '/?'
        if self.query is not None:
            self.url += "page=search&term={}&".format(self.query)
        self.url += "filter={}&sort={}&order={}".format(self.filter,self.sort, self.order)

    def get_html(self, headers={'User-Agent':'Firefox'}):
        req = rqst.Request(self.url, headers=headers)
        con = rqst.urlopen(req)
        self.soup = BeautifulSoup(con.read(), 'lxml')

    def set_url_offset(self, offset):
        if 'offset' in self.url:
            self.url = "&".join(self.url.split('&')[:-1] + ['offset={}'.format(offset)])
        else:
            self.url += "&offset={}".format(offset)

    def fetch(self):
        self.build_url()
        self.get_html()
        i = self.start_offset
        while True:
            self.set_url_offset(i)
            self.get_html()
            if self.soup.find('tr', {'class':'tlistrow'}):
                for x in self.soup.find_all('tr', {'class':'tlistrow'}):
                    yield Torrent(soup=x, color=self.color)
                if self.max_offset == -1: #fetch all pages
                    i += 1
                else:
                    if i < self.max_offset: i += 1
                    else: break
            else: #nothing found
                break


class Torrent:
    """
    The class representing a torrent entry.
    Required Arguments: (at least one required)
        - url (default = None)          the url pointing to the torrent page
        - soup (default = None)         the html BeautifulSoup block of the
                                        torrent page.
    """
    def __init__(self, url=None, soup=None, color=False):
        #connect to url and parse page
        if soup is None:
            if url[:5] != 'http:':
                self.url = 'http:' + self.url
            if 'showfiles' not in url:
                self.url = self.url + '&showfiles=1'
            else:
                self.url = url
            self.get_extended_infos()
        #directly parse page
        if soup is not None:
            self.soup = BeautifulSoup(soup, 'lxml')
            self.name  = self.soup.find('td', {'class':'tlistname'}).text
            #removes category from subcategory
            self.category, self.subcategory = self.soup.find('td', {'class':'tlisticon'}).find('a')['title'].split('>>')
            self.category, self.subcategory = self.category.strip(), self.subcategory.strip()
            self.subcategory = self.subcategory.replace(self.category, "")
            #url
            self.url = 'http:' + self.soup.find('td', {'class':'tlistname'}).find('a')['href']
            self.size = self.soup.find('td', {'class':'tlistsize'}).text
            #sets type according to <tlist class='xxx tlistrow'>
            self.type = {'tlistrow':'std', 'trusted':'trs',
                         'aplus':'a+', 'remake':'rmk'}[self.soup['class'][0]]
            try: #issue can happen if "status-unknown"
                self.seed = int(self.soup.find('td', {'class':'tlistsn'}).text)
                self.leech = int(self.soup.find('td', {'class':'tlistln'}).text)
            except AttributeError:
                self.seed, self.leech = 0,0
            self.downloads = int(self.soup.find('td', {'class':'tlistdn'}).text)
            self.messages = int(self.soup.find('td', {'class':'tlistmn'}).text)
            self.comments = []
        self.color = color

    def get_extended_infos(self, headers={'User-Agent':'Firefox'}):
        """Connect to url to fetch infos from torrent page instead of tlist entry"""
        self.get_html(headers=headers)
        self.name = self.soup.find('td', {'class':'viewtorrentname'}).text
        self.category, self.subcategory = self.soup.find('td', {'class':'viewcategory'}).text.split('>>')

        self.size = [x.text for x in self.soup.find_all('td', {'class':'vtop'}) if 'iB' in x.text][-1]
        self.seed = int(self.soup.find('span', {'class':'viewsn'}).text)
        self.leech = int(self.soup.find('span', {'class':'viewln'}).text)
        self.downloads = int(self.soup.find('span', {'class':'viewdn'}).text)
        self.comments = [Comment(x) for x in self.soup.find_all('div', {'class':'comment'})]
        self.messages = len(self.comments)
        #sets type according to <div class='xxx content'>
        self.type = {'content':'std', 'remake':'rmk', 'aplus':'a+','trusted':'trs'}\
                    [ self.soup.find('div',{'class':'content'})['class'][0] ]

    def get_html(self, headers={'User-Agent':'Firefox'}):
        req = rqst.Request(self.url, headers=headers)
        con = rqst.urlopen(req)
        self.soup = BeautifulSoup(con.read(), 'lxml')

    def __repr__(self):
        c = self.color
        typecolor = typecolor_dict[self.type]
        r =  ">> {}".format(typecolor(bold(self.name, c), c))
        r += "\n   - {} / {}".format(self.category, self.subcategory)
        r += "\n   - {}".format(self.size)
        r += "\n   - {}/{}".format(green('{} S'.format(self.seed), c),
                                    red('{} L'.format(self.leech), c))
        r += "\n   - {} DLs".format(self.downloads)
        r += "\n   - {}".format(self.messages)
        return r

    def download(self, directory=".", headers={'User-Agent' : 'Firefox'}):
        tid = re.search("tid=([0-9]*)", self.url).group(1)
        dl_url = "{}/?page=download&tid={}".format(URL, tid)
        d = Downloader(dl_url, directory, self.name+'.torrent', headers=headers)
        d.start()


class Comment:
    """
    The class representing a comment on a torrent page.
    Required Arguments:
        - comment_block             the html block <div class="comment">
                                    containing the comment
    """
    def __init__(self, comment_block):
        #search for date in comment under format YYYY-MM-DD at HH:MM
        datesearcher = re.search("([0-9]{4}-[0-9]{2}-[0-9]{2} at [0-9]{2}:[0-9]{2})",
                                 comment_block.text)
        self.date = dateutil.parser.parse(datesearcher.group(1))
        self.id = int(comment_block['id'][4:])
        self.user = comment_block.find('a').text
        self.content = comment_block.find('div',{'class':'cmain'}).text

    def __repr__(self):
        r =  "> Comment #{}\n".format(self.id)
        r += "> User '{}' {}:\n".format(self.user, self.date.isoformat())
        r += self.content
        return r


def parse_args(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("query", action='store', nargs='*', default=None,
                        help="The string to search for on Nyaa")
    parser.add_argument("-s", action="store", default="date",
                        choices=['date', 'seeders', 'leechers', 'downloads', 'size'],
                        help="the sorting key to use for the resulting torrent list")
    parser.add_argument("-o", action="store", default="desc",
                        choices=['asc', 'desc'],
                        help="the order to use for the resulting torrent list")
    parser.add_argument("-f", action="store", default='all',
                        choices=['all', 'noremake', 'trusted', 'aplus'],
                        help="the filter for the resulting torrent list")
    parser.add_argument("-r", action='store', default="1:1", metavar='range',
                        help="the ranges of pages to fetch")
    parser.add_argument('-c', '--color', action='store_true', default=False,
                        help="colored output (only in Unix terminals)")
    parser.add_argument('-q', '--quiet', action='store_true', default=False,
                        help='do not print any output')
    parser.add_argument('-d', '--detail', action='store_true', default=False,
                        help="print details about each torrent")
    parser.add_argument('-D', '--download', action='store_true', default=False,
                        help='download torrents files in current directory')
    #parse
    options = vars(parser.parse_args(args))
    #added special errors
    if ':' not in options['r']:
        parser.error("required range format is [start]:[stop]")
    start, end = " {} ".format(options['r']).split(':')
    options['start'] = 1 if start==" " else int(start)
    options['end'] = -1 if end==" " else int(end)
    if options['end'] != -1:
        if options['start'] > options['end']:
            parser.error("range start must be inferior to range stop")
    if options['color']:
        if os.name != 'posix':
            parser.error("colored output is only available on *nix systems")

    return options



#imports
__all__ = [__doc__, __version__, TList, Torrent, Comment, parse_args]

#main
def main():
    args = parse_args()
    for t in TList(query="".join(args['query']), sort=args['s'],
                   order=args['o'], filter=args['f'], color=args['color'],
                   start_page=args['start'], end_page=args['end']).fetch():
        if args['download']:
            t.download()
        if not args['quiet']:
            if not args['detail']:
                print(">", typecolor_dict[t.type](t.name, args['color']))
            else:
                print(t)

if __name__=='__main__':
    main()
